"""Generate JSONL with gene to phenotype/disease links"""

import gzip
import json
import typing

from google.protobuf.json_format import MessageToDict, ParseDict
import tqdm

from clinvar_data.pbs.clinvar_public_pb2 import (
    Allele,
    Assertion,
    ClassifiedRecord,
    VariationArchive,
)
from clinvar_data.pbs.extracted_vars import VersionedAccession
from clinvar_data.pbs.phenotype_link import GenePhenotypeRecord


def clean_omim(terms: typing.Set[str]) -> typing.Set[str]:
    return {term.split(".")[0] for term in terms}


def run_report(path_input: str, path_output: str, needs_hpo_terms: bool = True):  # noqa: C901
    """Read in file at path_input and generate link records to path_output."""

    if path_input.endswith(".gz"):
        inputf = gzip.open(path_input, "rt")
    else:
        inputf = open(path_input, "rt")

    if path_output.endswith(".gz"):
        outputf = gzip.open(path_output, "wt")
    else:
        outputf = open(path_output, "wt")

    with inputf, outputf:
        for line in tqdm.tqdm(inputf, desc="processing", unit=" JSONL records"):
            line_json = json.loads(line)
            variation_archive: VariationArchive = ParseDict(
                js_dict=line_json, message=VariationArchive()
            )

            vcv = VersionedAccession(
                accession=variation_archive.accession,
                version=variation_archive.version,
            )

            if not variation_archive.HasField("classified_record"):
                continue
            classified_record: ClassifiedRecord = variation_archive.classified_record
            if not variation_archive.classified_record.HasField("simple_allele"):
                continue
            simple_allele: Allele = classified_record.simple_allele
            hgnc_ids = [gene.hgnc_id for gene in simple_allele.genes if gene.HasField("hgnc_id")]

            for clinical_assertion in classified_record.clinical_assertions:
                scv = VersionedAccession(
                    accession=clinical_assertion.clinvar_accession.accession,
                    version=clinical_assertion.clinvar_accession.version,
                )
                germline_classification: str | None = None
                if (
                    clinical_assertion.HasField("classifications")
                    and clinical_assertion.classifications.HasField("germline_classification")
                    # and "pathogenic" in classification.germline_classification.lower()
                ):
                    germline_classification = (
                        clinical_assertion.classifications.germline_classification
                    )
                if not germline_classification:
                    continue

                if clinical_assertion.assertion not in (
                    Assertion.ASSERTION_VARIATION_TO_DISEASE,
                    Assertion.ASSERTION_VARIATION_TO_INCLUDED_DISEASE,
                ):
                    continue

                mondo_terms = []
                omim_terms = []
                hpo_terms = []
                for trait in clinical_assertion.trait_set.traits:
                    for xref in trait.xrefs:
                        if xref.db == "OMIM":
                            omim_terms.append(xref.id)
                        elif xref.db == "HP":
                            hpo_terms.append(xref.id)
                        elif xref.db == "MONDO":
                            mondo_terms.append(xref.id)

                for observed_in in clinical_assertion.observed_ins:
                    for trait in observed_in.trait_set.traits:
                        for xref in trait.xrefs:
                            if xref.db == "OMIM":
                                omim_terms.append(xref.id)
                            elif xref.db == "HP":
                                hpo_terms.append(xref.id)
                            elif xref.db == "MONDO":
                                mondo_terms.append(xref.id)

                mondo_terms = list(sorted(set(mondo_terms)))
                omim_terms = list(sorted(set(omim_terms)))
                hpo_terms = list(sorted(set(hpo_terms)))
                if not hpo_terms and needs_hpo_terms:
                    continue

                record = GenePhenotypeRecord(
                    vcv=vcv,
                    scv=scv,
                    germline_classification=germline_classification,
                    hgnc_ids=hgnc_ids,
                    mondo_terms=mondo_terms,
                    omim_terms=omim_terms,
                    hpo_terms=hpo_terms,
                )

                print(json.dumps(MessageToDict(record)), file=outputf)



"""Code to convert ClinVar XML to JSONL"""

import gzip
import json
import sys
import traceback
import typing

import click
from google.protobuf.json_format import MessageToDict
import tqdm
import xmltodict

from clinvar_data.conversion import dict_to_pb
from clinvar_data.conversion.normalizer import VariationArchiveNormalizer
from clinvar_data.pbs import clinvar_public

#: Total number of VariantArchive records in ClinVar on 2024-05-24: 2966486.
TOTAL_RECORDS: int = 3_000_000


def convert_variation_archive(json_va: dict) -> clinvar_public.VariationArchive:
    """Convert a ClinVarSet from JSON dict."""
    return dict_to_pb.ConvertVariationArchive.xmldict_data_to_pb(json_va)


def convert(
    input_file: str,
    output_file: str,
    max_records: int = 0,
    use_click: bool = False,
    show_progress: bool = True,
    fasta_ref_hg19: typing.Optional[str] = None,
    fasta_ref_hg38: typing.Optional[str] = None,
) -> int:
    """Run conversion from ClinVar XML to JSONL"""
    if input_file.endswith((".gz", ".bgz")):
        inputf: typing.Union[typing.BinaryIO, gzip.GzipFile] = gzip.open(input_file, "rb")
    elif use_click:
        inputf = click.open_file(input_file, "rb")
    else:
        inputf = open(input_file, "rb")

    if output_file.endswith(".gz"):
        outputf = gzip.open(output_file, "wt")
    elif use_click:
        outputf = click.open_file(output_file, "wt")
    else:
        outputf = open(output_file, "wt")

    pb: tqdm.tqdm | None = None
    if show_progress:
        pb = tqdm.tqdm(
            desc="parsing", unit=" VariationArchive records", smoothing=1.0, total=TOTAL_RECORDS
        )
    records_written = 0
    errors = 0

    normalizer = VariationArchiveNormalizer(fasta_ref_hg19, fasta_ref_hg38)

    def handle_variationarchive(ctx: list[dict], json_cvs: dict):
        """Handle single VariationArchive entry after parsing by ``xmltodict``."""
        try:
            key, val = ctx[-1]
            assert key == "VariationArchive"
            json_va = {key: {f"@{k}": v for k, v in val.items()}}
            json_va[key].update(json_cvs)
            data = convert_variation_archive(json_va)
            normalized_data_list = normalizer.normalize(data)
        except Exception:  # pragma: no cover
            nonlocal errors
            errors += 1
            print("Problem with data: exception and data follow", file=sys.stderr)
            traceback.print_exc()
            print(json_cvs, file=sys.stderr)
            return True

        for normalized_data in normalized_data_list:
            print(
                json.dumps(MessageToDict(normalized_data)),
                file=outputf,
            )

        nonlocal records_written
        records_written += 1
        if pb:
            pb.update(1)

        return max_records == 0 or records_written < max_records

    try:
        xmltodict.parse(inputf, item_depth=2, item_callback=handle_variationarchive)
    except xmltodict.ParsingInterrupted:  # pragma: no cover
        print(f"stopping after parsing {records_written} records", file=sys.stderr)

    if errors == 0:
        return 0
    else:
        print(f"a total of {errors} errors occurred", file=sys.stderr)
        return 1



"""Helper code for normalizing variants.

Note: one core assumption is that the contig names in the FASTA file
are prefixed with "chr", and the PAR on "chrY" is not N-masked, and
"chrMT" is the RCRS and present in the FASTA.

You will have to manually append a "chrMT" to "hg19.fa" and "hg38.fa".
"""

import typing

from logzero import logger
import pydantic
import pysam

from clinvar_data.conversion import normalize
from clinvar_data.pbs import clinvar_public
from clinvar_data.pbs.clinvar_public_pb2 import (
    Allele,
    ClassifiedRecord,
    VariationArchive,
)

#: Mapping from ambiguous IUPAC nucleotide codes to the possible nucleotides
IUPAC_AMBIGUOUS = {
    "R": "AG",
    "Y": "CT",
    "S": "GC",
    "W": "AT",
    "K": "GT",
    "M": "AC",
    "B": "CGT",
    "D": "AGT",
    "H": "ACT",
    "V": "ACG",
}


def is_ambiguous(seq: str) -> bool:
    """Check if a sequence contains ambiguous IUPAC nucleotide codes."""
    return any(key in seq for key in IUPAC_AMBIGUOUS.keys())


def expand_ambiguous(seq: str) -> typing.Iterator[str]:
    """Expand ambiguous IUPAC nucleotide codes in a sequence."""
    any_found = False
    for code, nts in IUPAC_AMBIGUOUS.items():
        try:
            pos = seq.index(code)
            any_found = True
            for nt in nts:
                yield from expand_ambiguous("".join((seq[:pos], nt, seq[pos + 1 :])))
        except ValueError:
            pass  # swallow
    if not any_found:
        yield seq


#: Chromosomes for which we have sequence.
CHROMS_WITH_SEQ = list(map(str, range(1, 23))) + ["X", "Y", "MT"]


class VcfVariant(pydantic.BaseModel):
    """VCF variant."""

    #: Chromosome.
    chrom: str
    #: Position.
    pos: int
    #: Reference allele.
    ref: str
    #: Alternate allele.
    alt: str

    model_config = pydantic.ConfigDict(frozen=True)


class DnaNormalizer:
    """Normalize a variant based on chrom/pos/ref/alt for DNA characters."""

    def __init__(self, fasta_ref: str):
        #: FASTA reference file.
        self.fasta_ref = pysam.FastaFile(fasta_ref)

    def normalize(self, variant: VcfVariant) -> VcfVariant:
        try:
            chrom, pos, ref, alt = normalize.normalize(
                self.fasta_ref, variant.chrom, variant.pos, variant.ref, variant.alt
            )
            if ref != variant.ref or alt != variant.alt:
                logger.info(
                    "normalized %s %s %s %s to %s %s %s %s",
                    variant.chrom,
                    variant.pos,
                    variant.ref,
                    variant.alt,
                    chrom,
                    pos,
                    ref,
                    alt,
                )
            return VcfVariant(chrom=chrom, pos=pos, ref=ref, alt=alt)
        except (normalize.RefEqualsAltError, normalize.WrongRefError) as e:
            logger.info("Skipping normalization because of error (will write out though): %s", e)
            return variant


class AmbiguousDnaNormalizer:
    """Normalize a variant, expanding ambiguous IUPAC codes in alternate allele.

    This implies normalizing to a list of variants.
    """

    def __init__(self, fasta_ref: str):
        #: Internal normalizer.
        self.normalizer = DnaNormalizer(fasta_ref)

    def normalize(self, variant: VcfVariant) -> typing.List[VcfVariant]:
        if is_ambiguous(variant.alt):
            return [
                self.normalizer.normalize(variant.model_copy(update={"alt": alt}))
                for alt in expand_ambiguous(variant.alt)
            ]
        else:
            return [self.normalizer.normalize(variant)]


def vcf_variant_from_variant_archive(
    va: VariationArchive, assembly: typing.Literal["GRCh37", "GRCh38"]
) -> typing.Optional[VcfVariant]:
    """Extract ``VcfVariant`` from a ``VariationArchive`mak` (via sequence location)."""
    if not va.HasField("classified_record"):
        return None
    classified_record: ClassifiedRecord = va.classified_record
    if not classified_record.HasField("simple_allele"):
        return None
    simple_allele: Allele = classified_record.simple_allele
    for location in simple_allele.locations or []:
        for sequence_location in location.sequence_locations or []:
            if (
                sequence_location.assembly == assembly
                and sequence_location.HasField("reference_allele_vcf")
                and sequence_location.HasField("alternate_allele_vcf")
            ):
                chrom = clinvar_public.Chromosome.Name(sequence_location.chr)[len("CHROMOSOME_") :]
                if chrom in CHROMS_WITH_SEQ:
                    return VcfVariant(
                        chrom=f"chr{chrom}",
                        pos=sequence_location.position_vcf,
                        ref=sequence_location.reference_allele_vcf,
                        alt=sequence_location.alternate_allele_vcf,
                    )
    return None


def write_vcf_variant_to_va(
    va: VariationArchive, vcf_variant: VcfVariant, assembly: typing.Literal["GRCh37", "GRCh38"]
):
    """Write VCF variant coordinates to a ``VariationArchive``."""
    if not va.HasField("classified_record"):
        return None
    classified_record: ClassifiedRecord = va.classified_record
    if not classified_record.HasField("simple_allele"):
        return None
    simple_allele: Allele = classified_record.simple_allele
    for location in simple_allele.locations or []:
        for sequence_location in location.sequence_locations or []:
            if (
                sequence_location.assembly == assembly
                and sequence_location.HasField("reference_allele_vcf")
                and sequence_location.HasField("alternate_allele_vcf")
            ):
                sequence_location.position_vcf = vcf_variant.pos
                sequence_location.reference_allele_vcf = vcf_variant.ref
                sequence_location.alternate_allele_vcf = vcf_variant.alt


class MatchedVcfVariants(pydantic.BaseModel):
    """Stores pair of matched hg19/hg38 variants."""

    #: HG19 variant.
    hg19: typing.Optional[VcfVariant] = None
    #: HG38 variant.
    hg38: typing.Optional[VcfVariant] = None


class VariationArchiveNormalizer:
    """Normalize the VCF variant description within a ``VariationArchive``.

    This will normalize the variant for both the GRCh37 and GRCh38 assemblies.
    Note that this will also expand ambiguous IUPAC codes in the alternate allele,
    leading to duplicate records in terms of VCV but still unique in terms
    of VCF positions.
    """

    def __init__(
        self,
        fasta_ref_hg19: typing.Optional[str] = None,
        fasta_ref_hg38: typing.Optional[str] = None,
    ):
        #: Normalizer to use for GRCh37.
        self.normalizer_hg19 = AmbiguousDnaNormalizer(fasta_ref_hg19) if fasta_ref_hg19 else None
        #: Normalizer to use for GRCh38.
        self.normalizer_hg38 = AmbiguousDnaNormalizer(fasta_ref_hg38) if fasta_ref_hg38 else None

    def normalize(self, va: VariationArchive) -> typing.List[VariationArchive]:
        """Normalize the VCF variant in ``va``.

        The algorithm is as follows:

        1. extract VCF-style coordinates on both GRCh37 and GRCh38 from ``va``
        2. normalize them, expanding ambiguous IUPAC codes in the alternate allele
        3. match the resulting variants based on expanded non-IUPAC alt alleles
        4. for each matched pair, create a copy of the original ``va``, write
           the normalized VCF variant back to the copy
        """
        # extract
        vcf_hg19 = vcf_variant_from_variant_archive(va, "GRCh37")
        vcf_hg38 = vcf_variant_from_variant_archive(va, "GRCh38")
        if vcf_hg19 is None and vcf_hg38 is None:
            return [va]
        # normalize
        vcfs_hg19 = (
            self.normalizer_hg19.normalize(vcf_hg19)
            if self.normalizer_hg19 and vcf_hg19
            else ([vcf_hg19] if vcf_hg19 is not None else [])
        )
        vcfs_hg38 = (
            self.normalizer_hg38.normalize(vcf_hg38)
            if self.normalizer_hg38 and vcf_hg38
            else ([vcf_hg38] if vcf_hg38 is not None else [])
        )
        # short-circuit if nothing to normalize
        if vcfs_hg19 == [vcf_hg19] and vcfs_hg38 == [vcf_hg38]:
            return [va]
        # match
        matched_vcfs: dict[str, MatchedVcfVariants] = {}
        for vcf_variant in vcfs_hg19:
            matched_vcfs[vcf_variant.alt] = MatchedVcfVariants(hg19=vcf_variant)
        for vcf_variant in vcfs_hg38:
            if vcf_variant.alt in matched_vcfs:
                matched_vcfs[vcf_variant.alt].hg38 = vcf_variant
            else:
                matched_vcfs[vcf_variant.alt] = MatchedVcfVariants(hg38=vcf_variant)
        return [
            self._create_copy_with_variant(va, matched_vcf) for matched_vcf in matched_vcfs.values()
        ]

    def _create_copy_with_variant(
        self, va: VariationArchive, matched_vcf: MatchedVcfVariants
    ) -> VariationArchive:
        result = VariationArchive()
        result.MergeFrom(va)
        if matched_vcf.hg19:
            write_vcf_variant_to_va(result, matched_vcf.hg19, "GRCh37")
        if matched_vcf.hg38:
            write_vcf_variant_to_va(result, matched_vcf.hg38, "GRCh38")
        return result



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: clinvar_data/pbs/gene_impact.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from clinvar_data.pbs import (
    clinvar_public_pb2 as clinvar__data_dot_pbs_dot_clinvar__public__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n"clinvar_data/pbs/gene_impact.proto\x12\x1c\x63linvar_data.pbs.gene_impact\x1a%clinvar_data/pbs/clinvar_public.proto"\xdb\x02\n\x10GeneImpactCounts\x12\x0f\n\x07hgnc_id\x18\x01 \x01(\t\x12R\n\rimpact_counts\x18\x02 \x03(\x0b\x32;.clinvar_data.pbs.gene_impact.GeneImpactCounts.ImpactCounts\x1a\xe1\x01\n\x0cImpactCounts\x12=\n\x0bgene_impact\x18\x01 \x01(\x0e\x32(.clinvar_data.pbs.gene_impact.GeneImpact\x12\x14\n\x0c\x63ount_benign\x18\x02 \x01(\r\x12\x1b\n\x13\x63ount_likely_benign\x18\x03 \x01(\r\x12$\n\x1c\x63ount_uncertain_significance\x18\x04 \x01(\r\x12\x1f\n\x17\x63ount_likely_pathogenic\x18\x05 \x01(\r\x12\x18\n\x10\x63ount_pathogenic\x18\x06 \x01(\r*\xf9\x04\n\nGeneImpact\x12\x1b\n\x17GENE_IMPACT_UNSPECIFIED\x10\x00\x12\'\n#GENE_IMPACT_THREE_PRIME_UTR_VARIANT\x10\x01\x12&\n"GENE_IMPACT_FIVE_PRIME_UTR_VARIANT\x10\x02\x12-\n)GENE_IMPACT_DOWNSTREAM_TRANSCRIPT_VARIANT\x10\x03\x12"\n\x1eGENE_IMPACT_FRAMESHIFT_VARIANT\x10\x04\x12\x1d\n\x19GENE_IMPACT_INFRAME_INDEL\x10\x05\x12\x1a\n\x16GENE_IMPACT_START_LOST\x10\x06\x12\x1e\n\x1aGENE_IMPACT_INTRON_VARIANT\x10\x07\x12 \n\x1cGENE_IMPACT_MISSENSE_VARIANT\x10\x08\x12-\n)GENE_IMPACT_NON_CODING_TRANSCRIPT_VARIANT\x10\t\x12\x1b\n\x17GENE_IMPACT_STOP_GAINED\x10\n\x12&\n"GENE_IMPACT_NO_SEQUENCE_ALTERATION\x10\x0b\x12\'\n#GENE_IMPACT_SPLICE_ACCEPTOR_VARIANT\x10\x0c\x12$\n GENE_IMPACT_SPLICE_DONOR_VARIANT\x10\r\x12\x19\n\x15GENE_IMPACT_STOP_LOST\x10\x0e\x12"\n\x1eGENE_IMPACT_SYNONYMOUS_VARIANT\x10\x0f\x12+\n\'GENE_IMPACT_UPSTREAM_TRANSCRIPT_VARIANT\x10\x10*\xd6\x02\n\x14\x43linicalSignificance\x12%\n!CLINICAL_SIGNIFICANCE_UNSPECIFIED\x10\x00\x12 \n\x1c\x43LINICAL_SIGNIFICANCE_BENIGN\x10\x01\x12\'\n#CLINICAL_SIGNIFICANCE_LIKELY_BENIGN\x10\x02\x12\x30\n,CLINICAL_SIGNIFICANCE_UNCERTAIN_SIGNIFICANCE\x10\x03\x12+\n\'CLINICAL_SIGNIFICANCE_LIKELY_PATHOGENIC\x10\x04\x12$\n CLINICAL_SIGNIFICANCE_PATHOGENIC\x10\x05\x12&\n"CLINICAL_SIGNIFICANCE_NOT_PROVIDED\x10\x06\x12\x1f\n\x1b\x43LINICAL_SIGNIFICANCE_OTHER\x10\x07\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "clinvar_data.pbs.gene_impact_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_GENEIMPACT"]._serialized_start = 458
    _globals["_GENEIMPACT"]._serialized_end = 1091
    _globals["_CLINICALSIGNIFICANCE"]._serialized_start = 1094
    _globals["_CLINICALSIGNIFICANCE"]._serialized_end = 1436
    _globals["_GENEIMPACTCOUNTS"]._serialized_start = 108
    _globals["_GENEIMPACTCOUNTS"]._serialized_end = 455
    _globals["_GENEIMPACTCOUNTS_IMPACTCOUNTS"]._serialized_start = 230
    _globals["_GENEIMPACTCOUNTS_IMPACTCOUNTS"]._serialized_end = 455
# @@protoc_insertion_point(module_scope)



"""Determin ACMG class by frequency."""

import gzip
import json
import typing

from google.protobuf.json_format import MessageToDict, Parse
import tqdm

from clinvar_data.pbs.class_by_freq import CoarseClinicalSignificance
from clinvar_data.pbs.class_by_freq_pb2 import GeneCoarseClinsigFrequencyCounts
from clinvar_data.pbs.clinvar_public_pb2 import VariationArchive

#: Default thresholds for frequency.
DEFAULT_THRESHOLDS = [
    0.0,
    1e-05,
    2.5e-05,
    5e-05,
    0.0001,
    0.00025,
    0.0005,
    0.001,
    0.0025,
    0.005,
    0.01,
    0.025,
    0.05,
    0.1,
    0.25,
    0.5,
    1.0,
]


class ConvertCoarseClinicalSignificance:
    """Static method helper for germline clinical significance summary string to coarse enum conversion."""

    #: Dict for conversion.
    CONVERT: dict[str, CoarseClinicalSignificance.ValueType] = {
        "Pathogenic": CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_PATHOGENIC,
        "Pathogenic/Likely pathogenic": CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_PATHOGENIC,
        "Likely pathogenic": CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_PATHOGENIC,
        "Uncertain significance": CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_UNCERTAIN,
        "Likely benign": CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_BENIGN,
        "Benign": CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_BENIGN,
        "Benign/Likely benign": CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_BENIGN,
    }

    @classmethod
    def from_str(cls, string_value: str) -> CoarseClinicalSignificance.ValueType:
        """Convert string to protobuf enum value."""
        return cls.CONVERT.get(
            string_value, CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_UNSPECIFIED
        )


def zero_counts(count: int):
    return {
        klass.number: [0] * (count + 1)
        for klass in CoarseClinicalSignificance.DESCRIPTOR.values
        if klass.number != CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_UNSPECIFIED
    }


def locate(thresholds: typing.List[float], value: typing.Optional[float]) -> int:
    """Locate the index of the threshold that the value falls into."""
    if value is None:
        return 0
    for i, threshold in enumerate(thresholds):
        if value <= threshold:
            return i
    return len(thresholds)


def generate_counts(path_input: str, thresholds: typing.List[float]):
    """Generate counts from variant JSONL file."""
    counts = {}

    if path_input.endswith(".gz"):
        inputf = gzip.open(path_input, "rt")
    else:
        inputf = open(path_input, "rt")

    with inputf:
        for line in tqdm.tqdm(inputf, desc="processing", unit=" JSONL records"):
            va: VariationArchive = Parse(line, message=VariationArchive())

            # Obtain germline classification description or skip record if has none.
            if not va.classified_record.HasField("classifications"):
                continue
            elif not va.classified_record.classifications.HasField("germline_classification"):
                continue
            elif not va.classified_record.classifications.germline_classification.HasField(
                "description"
            ):
                continue
            else:
                description: str = (
                    va.classified_record.classifications.germline_classification.description
                )
                pathogenicity: CoarseClinicalSignificance.ValueType = (
                    ConvertCoarseClinicalSignificance.from_str(description)
                )
            if pathogenicity == CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_UNSPECIFIED:
                continue

            # Obtain minor allele frequency.
            gmaf: float | None = None
            if not va.HasField("classified_record"):
                continue
            elif not va.classified_record.HasField("simple_allele"):
                continue
            elif va.classified_record.simple_allele.HasField("global_minor_allele_frequency"):
                gmaf = va.classified_record.simple_allele.global_minor_allele_frequency.value

            # Try to get VCF location and skip unless is sequence variant.
            is_seqvar = False
            simple_allele = va.classified_record.simple_allele
            for location in simple_allele.locations or []:
                for sequence_location in location.sequence_locations or []:
                    if (
                        sequence_location.HasField("reference_allele")
                        and len(sequence_location.reference_allele) < 50
                        and sequence_location.HasField("alternate_allele")
                        and len(sequence_location.alternate_allele) < 50
                    ) or (
                        sequence_location.HasField("reference_allele_vcf")
                        and len(sequence_location.reference_allele_vcf) < 50
                        and sequence_location.HasField("alternate_allele_vcf")
                        and len(sequence_location.alternate_allele_vcf) < 50
                    ):
                        is_seqvar = True
                        break
            if not is_seqvar:
                continue

            # Obtain gene HGNC ID.
            hgnc_id: str | None = None
            for gene in va.classified_record.simple_allele.genes:
                if not gene.HasField("hgnc_id"):
                    continue
                else:
                    hgnc_id = gene.hgnc_id
                    break
            if not hgnc_id:
                continue

            if hgnc_id not in counts:
                counts[hgnc_id] = zero_counts(len(thresholds))

            idx = locate(thresholds, gmaf)
            counts[hgnc_id][pathogenicity][idx] += 1

    return counts


def write_report(counts: dict, path_output: str):
    """Write counts report to output."""
    if path_output.endswith(".gz"):
        outputf = gzip.open(path_output, "wt")
    else:
        outputf = open(path_output, "wt")

    with outputf:
        for hgnc_id, counts in sorted(counts.items(), key=lambda x: int(x[0][5:])):
            record: GeneCoarseClinsigFrequencyCounts = GeneCoarseClinsigFrequencyCounts(
                hgnc_id=hgnc_id,
                pathogenic_counts=counts[
                    CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_PATHOGENIC
                ],
                uncertain_counts=counts[
                    CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_UNCERTAIN
                ],
                benign_counts=counts[
                    CoarseClinicalSignificance.COARSE_CLINICAL_SIGNIFICANCE_BENIGN
                ],
            )
            print(
                json.dumps(MessageToDict(record)),
                file=outputf,
            )


def run_report(path_input: str, path_output: str, thresholds: typing.List[float]):
    counts = generate_counts(path_input, thresholds)
    write_report(counts, path_output)



#!/usr/bin/env python

"""
This script is a python implementation of the algorithm for variant
normalization described by Tan et al 2015:

Tan A, Abecasis GR, Kang HM. Unified representation of genetic variants.
Bioinformatics. 2015 Jul 1;31(13):2202-4. doi: 10.1093/bioinformatics/btv112.
Epub 2015 Feb 19. PubMed PMID: 25701572.

The authors have made a C++ implementation available in vt as vt normalize
And the source code is viewable here: https://github.com/atks/vt

For our purposes, we wanted a Python implementation so that we could
build end-to-end scripts in Python.

If you use this, please cite Tan et al 2015.

A note about when this is useful. In VCFs generated with GATK (or probably
other tools) from short read sequencing data, variants are already left-aligned
but may be non-minimal to the extent that indels overlap with other variants.
For those cases, minimal_representation.py is sufficient to convert variants
to minimal representation. However, genomic coordinates converted from HGVS
(we have encountered this when parsing the ClinVar XML dump) may be not only
non-minimal but also right-aligned rather than left-aligned, and may contain
hyphens. For those situations, use this script (or just run vt normalize).

Usage: normalize.py -R $b37ref < bad_file.txt > good_file.txt
"""


import typing

import pysam


class RefEqualsAltError(Exception):
    """
    An Error class for rare cases where REF == ALT (seen in ClinVar XML)
    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class InvalidNucleotideSequenceError(Exception):
    """
    An Error class for REF or ALT values that are not valid nucleotide sequences
    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class WrongRefError(Exception):
    """
    An Error class for variants where the REF does not match the reference genome
    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


def normalize(
    pysam_fasta: pysam.FastaFile, chrom: str, pos: int, ref: str, alt: str
) -> typing.Tuple[str, int, str, str]:
    """
    Accepts a pysam FastaFile object pointing to the reference genome, and
    chrom, pos, ref, alt genomic coordinates, and normalizes them.
    """
    pos = int(pos)  # make sure position is an integer
    ref = ref.upper()
    alt = alt.upper()
    # Remove variants that contain invalid nucleotides
    if any(letter not in ["A", "C", "G", "T", "N", "-"] for letter in ref + alt):
        raise InvalidNucleotideSequenceError(
            "Invalid nucleotide sequence: %s %s %s %s" % (chrom, pos, ref, alt)
        )
    # use blanks instead of hyphens
    if ref == "-":
        ref = ""
    if alt == "-":
        alt = ""
    # check whether the REF is correct
    true_ref = pysam_fasta.fetch(chrom, pos - 1, pos - 1 + len(ref)).upper()
    if ref != true_ref:
        raise WrongRefError(
            "Incorrect REF value: %s %s %s %s (actual REF should be %s)"
            % (chrom, pos, ref, alt, true_ref)
        )
    # Prevent infinite loops in cases where REF == ALT.
    # We have encountered this error in genomic coordinates from the ClinVar XML file
    if ref == alt:
        raise RefEqualsAltError(
            "The REF and ALT allele are the same: %s %s %s %s" % (chrom, pos, ref, alt)
        )
    # Time-saving shortcut for SNPs that are already minimally represented
    if (
        len(ref) == 1
        and len(alt) == 1
        and ref in ["A", "C", "G", "T"]
        and alt in ["A", "C", "G", "T"]
    ):
        return chrom, pos, ref, alt
    # This first loop left-aligns and removes excess nucleotides on the right.
    # This is Algorithm 1 lines 1-6 from Tan et al 2015
    keep_working = True
    while keep_working:
        keep_working = False
        if len(ref) > 0 and len(alt) > 0 and ref[-1] == alt[-1]:
            ref = ref[:-1]
            alt = alt[:-1]
            keep_working = True
        if len(ref) == 0 or len(alt) == 0:
            preceding_base = pysam_fasta.fetch(chrom, pos - 2, pos - 1)
            ref = preceding_base + ref
            alt = preceding_base + alt
            pos = pos - 1
            keep_working = True
    # This second loop removes excess nucleotides on the left. This is Algorithm 1 lines 7-8.
    while len(ref) > 1 and len(alt) > 1 and ref[0] == alt[0]:
        ref = ref[1:]
        alt = alt[1:]
        pos = pos + 1
    return chrom, pos, ref, alt
